"""
PyMBS is a Python library for use in modeling Mortgage-Backed Securities.

Copyright (C) 2019  Brian Farrell

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact: brian.farrell@me.com
"""

from collections import namedtuple
import decimal
import re

import pandas as pd

from pymbs.config import config
from pymbs.exceptions import PayRuleError


dec = decimal.Decimal

PNT = namedtuple('Pandas', [
    'Index',
    'period',
    'payment_date',
    'beginning_balance',
    'smm',
    'scheduled_payment',
    'net_interest',
    'scheduled_principal',
    'prepayment',
    'total_principal',
    'cash_flow',
    'ending_balance',
    'buckets'
])


def _month_range(start, periods=360):
    """
    https://stackoverflow.com/a/40040932
    """
    rng = pd.date_range(pd.Timestamp(start) - pd.offsets.MonthBegin(),
                        periods=periods,
                        freq='MS')
    ret = (rng + pd.offsets.Day(pd.Timestamp(start).day - 1)).to_series()
    ret.loc[ret.dt.month > rng.month] -= pd.offsets.MonthEnd(1)
    return pd.DatetimeIndex(ret)


def _round_dec(decimal_value, precision=config.round_precision):
    places = dec('10') ** dec(f"{-precision}")
    rounded_value = decimal_value.quantize(places)

    return rounded_value


def _parse_waterfall(waterfall):
    wf = []
    for rule in waterfall:
        pr = re.match(
            r'(?P<full_rule>(?P<pay_rule>\w+)\('
            r'(?P<buckets>\[(?:\'\w+\'(?:,\s|\]))*)'
            r'(?:,\s(?P<tranches>\[(?:\'\w+\'(?:,\s|\]))*))?(?:(?:,\s)'
            r'(?:\'(?P<expression>.*)\'))?\))',
            rule,
            re.S
        )

        if not pr:
            raise PayRuleError(rule, 'no_parse')

        if pr.group('pay_rule') == 'pay_accrue':
            if pr.group('tranches') or pr.group('expression'):
                raise PayRuleError(pr.group('full_rule'), 'invalid')
            else:
                func = pr.group('full_rule')
                # In the case of the pay_accrue rule, the tranches to accrue
                # are captured by the 'buckets' group
                tranches = pr.group('buckets')
        elif pr.group('pay_rule') == 'calculate':
            if pr.group('tranches'):
                raise PayRuleError(pr.group('full_rule'), 'invalid')
            else:
                func = pr.group('full_rule')
                tranches = None
        else:
            if pr.group('pay_rule') not in ['pay_accrue', 'calculate']:
                if pr.group('expression') or not \
                        (pr.group('buckets') and pr.group('tranches')):
                    raise PayRuleError(pr.group('full_rule'), 'invalid')
                else:
                    func = pr.group('full_rule')
                    tranches = pr.group('tranches')
        wf.append(dict(tranches=tranches, func=func))
    return wf


def _parse_expression(exp):
    expr_tokens = re.match(
        r'(?P<first_term>(?:COLL|(?:^COLL|[A-Za-z0-9]))+?_'
        r'(?:PRINCIPAL|ACCRUAL|PREPAY|BALANCE|SCHEDULED|VAR))'
        r'(?:\s(?P<operator>\+|\-|\*|\/)\s)'
        r'(?=(?:(?:COLL|(?:^COLL|[A-Za-z0-9])+)_'
        r'(?:PRINCIPAL|ACCRUAL|PREPAY|BALANCE|SCHEDULED|VAR)))'
        r'(?<=(?:\s(?:\+|\-|\*|\/)\s))'
        r'(?P<second_term>(?:COLL|(?:^COLL|[A-Za-z0-9]))+?_'
        r'(?:PRINCIPAL|ACCRUAL|PREPAY|BALANCE|SCHEDULED|VAR))', exp, re.S)

    first_term_tokens = re.findall(
        r'(?P<token>[A-Za-z0-9]+)', expr_tokens.group('first_term'), re.S)

    second_term_tokens = re.findall(
        r'(?P<token>[A-Za-z0-9]+)', expr_tokens.group('second_term'), re.S)

    # If the first token is 'COLL', then we assume that it is referring to the
    # collater cash flow passed to the calling function
    if first_term_tokens[0] == 'COLL':
        ft_1 = 'payment'
    # If the first token is NOT COLL and the second token is NOT VAR, then we
    # assume that it is referreing to cash flow of a Tranche in the same
    # group as the one for which the waterfall is being run
    if first_term_tokens[0] != 'COLL' and first_term_tokens[1] != 'VAR':
        ft_1 = (f"model['groups'][group_id]['tranches']"
                f"['{first_term_tokens[0]}'].periodic_cf")

    # If the second token is VAR, then the first token is the name of the
    # variable being evaluated
    if first_term_tokens[1] == 'VAR':
        ft_1 = first_term_tokens[0]
        ft_2 = ''
    elif first_term_tokens[1] == 'PRINCIPAL':
        ft_2 = 'total_principal'
    elif first_term_tokens[1] == 'ACCRUAL':
        ft_2 = 'accrual'
    elif first_term_tokens[1] == 'PREPAY':
        ft_2 = 'prepayment'
    elif first_term_tokens[1] == 'BALANCE':
        ft_2 = 'beginning_balance'
    elif first_term_tokens[1] == 'SCHEDULED':
        ft_2 = 'scheduled_principal'

    # If the first token is 'COLL', then we assume that it is referring to the
    # collater cash flow passed to the calling function
    if second_term_tokens[0] == 'COLL':
        st_1 = 'payment'
    # If the first token is NOT COLL and the second token is NOT VAR, then we
    # assume that it is referreing to cash flow of a Tranche in the same
    # group as the one for which the waterfall is being run
    if second_term_tokens[0] != 'COLL' and first_term_tokens[1] != 'VAR':
        st_1 = (f"model['groups'][group_id]['tranches']"
                f"['{second_term_tokens[0]}'].periodic_cf")

    # If the second token is VAR, then the first token is the name of the
    # variable being evaluated
    if second_term_tokens[1] == 'VAR':
        st_1 = second_term_tokens[0]
        st_2 = ''
    elif second_term_tokens[1] == 'PRINCIPAL':
        st_2 = 'total_principal'
    elif second_term_tokens[1] == 'ACCRUAL':
        st_2 = 'accrual'
    elif second_term_tokens[1] == 'PREPAY':
        st_2 = 'prepayment'
    elif second_term_tokens[1] == 'BALANCE':
        st_2 = 'beginning_balance'
    elif second_term_tokens[1] == 'SCHEDULED':
        st_2 = 'scheduled_principal'

    if ft_2 == '':
        term_1 = f"{ft_1} "
    else:
        term_1 = f"{ft_1}.{ft_2}"

    if st_2 == '':
        term_2 = f"{st_1} "
    else:
        term_2 = f"{st_1}['{st_2}']"

    operator = f" {expr_tokens.group('operator')} "

    return f"{term_1}{operator}{term_2}"



"""
PyMBS is a Python library for use in modeling Mortgage-Backed Securities.

Copyright (C) 2019  Brian Farrell

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact: brian.farrell@me.com
"""

no_deal = (
    "\nNo deal has been loaded yet. "
    "Please load a deal before loading a model."
)

no_model = (
    "\nNo model has been loaded yet. "
    "Please load a model before continuing."
)


def _no_config_msg(pymbs_config_path, help_url):
    no_config_msg = (
        f"No pymbs configuration file was found at "
        f"{pymbs_config_path}"
        f"\nA blank file has been located at that location."
        f"\n\nPlease see {help_url}"
        f"\nfor information on how to edit the file before proceeding.\n"
    )
    return no_config_msg


class PrepaymentBenchmarkError(Exception):
    """Exception raised for errors in the input.

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, prepayment_benchmark, message):
        self.prepayment_benchmark = prepayment_benchmark
        self.message = message


class DateError(Exception):
    """Exception raised for errors in the input.

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, date, message):
        self.date = date
        self.message = message


class PayRuleError(Exception):
    """Exception raised for errors in the input.

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, pay_rule, error):
        self.pay_rule = pay_rule
        self.error = error
        if self.error == 'invalid':
            self.message = f"The pay rule {pay_rule} is not a valid pay rule."
        elif self.error == 'no_parse':
            self.message = (
                f"The pay rule | {pay_rule} | was not parsed successfully. "
                f"\n\t\tPlease check the syntax."
            )

    def __str__(self):
        return self.message



"""
PyMBS is a Python library for use in modeling Mortgage-Backed Securities.
Copyright (C) 2019  Brian Farrell

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact: brian.farrell@me.com
"""

import pytest

from pymbs.api import NotFoundError
from pymbs.enums import iTunesESrA


def test_get_track_raises(app):
    """get_track should raise an exception when track can't be found.

    The calls to app.get_playlist() and app.get_tracks() is what
    distinguishes this test as a Functional test. For the purposes of
    this test, we are assuming that theses methods return a valid
    iTunes objects.

    The target method under test here is app.get_track()
    """
    requested_playlist = 'Music'
    playlist = app.get_playlist(requested_playlist)
    collection = app.get_tracks(playlist)
    requested_track = "Weekend At Bernie's XLII"

    with pytest.raises(NotFoundError):
        app.get_track(collection, requested_track)


def test_track_in(app):
    requested_playlist = 'Music'
    playlist = app.get_playlist(requested_playlist)
    collection = app.get_tracks(playlist)
    requested_track = "A Salty Dog"

    track = app.get_track(collection, requested_track)
    assert track.name() == requested_track


def test_search_track(app):
    playlist = 'Music'
    search_text = "Big Yellow Taxi"
    constraint = iTunesESrA.SONGS

    results = app.search(playlist, search_text, constraint)

    for track in results:
        print(track.name())

    assert len(results) > 0



"""
PyMBS is a Python library for use in modeling Mortgage-Backed Securities.
Copyright (C) 2019  Brian Farrell

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact: brian.farrell@me.com
"""

"""
Subcutaneous Test Namespace
"""



"""
PyMBS is a Python library for use in modeling Mortgage-Backed Securities.

Copyright (C) 2019  Brian Farrell

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact: brian.farrell@me.com
"""

from copy import copy
import decimal

from pymbs.config import config
from pymbs.utils import _parse_expression, PNT

dec = decimal.Decimal

d0 = dec('0')
cleanup = dec('1E-2')


def calculate(bucket, exp):
    model = config.cache['model']  # noqa
    group_id = config.cache['group_id']  # noqa
    prepay_scenario = config.cache['prepay_scenario']  # noqa
    payment = config.cache['payment']  # noqa

    expression = _parse_expression(exp)
    b = eval(expression)
    if 'buckets' in payment._fields:
        if payment.buckets[bucket]:
            payment.buckets[bucket] += b
        else:
            payment.buckets[bucket] = b
    else:
        this_bucket = f"{bucket[0]}"
        buckets = {this_bucket: b}
        payment = PNT(*list(payment), buckets)
        config.cache['payment'] = payment


def pay_accrual(bucket, tranche):
    pass


def pay_accrue(tranche_ids):
    tranches = config.cache['tranches']
    for tranche in tranches:
        tranche.pay_accrue()


def pay_pro_rata(bucket, tranches):
    ctx = decimal.getcontext()
    ctx.prec = 15
    ctx.Emax = 10
    ctx.Emin = -9
    decimal.setcontext(ctx)

    model = config.cache['model']  # noqa
    group_id = config.cache['group_id']  # noqa
    prepay_scenario = config.cache['prepay_scenario']  # noqa
    payment = config.cache['payment']
    current_payment = ctx.create_decimal(payment.buckets[bucket[0]])
    orig_current_payment = copy(current_payment)

    target_tranches = []
    for tranche_id in tranches:
        t = model['groups'][group_id]['tranches'][tranche_id]
        target_tranches.append(t)

    upb_sum = d0
    for tranche in target_tranches:
        upb_sum += tranche.upb

    if upb_sum > d0:
        for tranche in target_tranches:
            tranche.pro_rated_ratio = tranche.upb / upb_sum

    if (current_payment > d0 and upb_sum > d0):
        for tranche in target_tranches:
            projected_payment = \
                orig_current_payment * tranche.pro_rated_ratio
            actual_payment = min(
                projected_payment,
                tranche.upb,
                current_payment,
            )
            tranche.pay_principal(actual_payment)
            current_payment -= actual_payment
            if current_payment < cleanup:
                current_payment = d0
            payment.buckets[bucket[0]] = current_payment


def pay_sequential(bucket, tranches):
    ctx = decimal.getcontext()
    ctx.prec = 15
    ctx.Emax = 10
    ctx.Emin = -9
    decimal.setcontext(ctx)

    model = config.cache['model']  # noqa
    group_id = config.cache['group_id']  # noqa
    prepay_scenario = config.cache['prepay_scenario']  # noqa
    payment = config.cache['payment']
    current_payment = ctx.create_decimal(payment.buckets[bucket[0]])

    target_tranches = []
    for tranche_id in tranches:
        t = model['groups'][group_id]['tranches'][tranche_id]
        target_tranches.append(t)

    for tranche in target_tranches:
        if (current_payment > d0 and tranche.upb > d0):
            actual_payment = min(
                current_payment,
                tranche.upb,
            )
            tranche.pay_principal(actual_payment)
            current_payment -= actual_payment
            if current_payment < cleanup:
                current_payment = d0
            payment.buckets[bucket[0]] = current_payment


def pay_concurrent(bucket, tranches, ratios):
    pass



"""
PyMBS is a Python library for use in modeling Mortgage-Backed Securities.

Copyright (C) 2019  Brian Farrell

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact: brian.farrell@me.com
"""

import decimal
import json
import os
from typing import Optional, TextIO

import pandas as pd

from pymbs import core, exceptions
from pymbs.config import config
from pymbs.enums import ExitCode
from pymbs.tranche import IndexRate, Tranche

__all__ = ['load_deal', 'load_assumed_collat', 'load_model', 'run_collat_cf',
           'show_wals', 'load_tranches', 'disperse_cf']


pd.set_option('display.max_rows', config.max_rows)
pd.set_option('precision', config.precision)


def load_deal(series_name: str) -> dict:
    """Load the Terms Sheet for the REMIC Series name provided.

    The Terms Sheet is essential for modeling the deal. If this function is
    not called at the start, most of the other functions will fail.

    Within the project_dir, specified in the user's configuration file, it is
    expected that there will be one subdirectory for each REMIC Series name.
    Inside the Series directory, it is expected that the Terms Sheet will be
    a file named ``series_ts.json``, where 'series' refers to the actual
    Series name.

    Args:
        series_name: The REMIC Series name. Note that this value is passed
        as a string, so it need not be strictly an integer value, i.e. 'T-074'.
        The value *must* be the same as the name of the deal directory, which
        is a subdirectory of the project_directory.

    Returns:
        A dictionary object, representing the Terms Sheet.
        If the Terms Sheet file is not located, returns Exit Code 66.

    Example:
        Given: REMIC Series Name is ``'2618'``

        Path: ``/project_dir/2618/2618_ts.json``
    """
    ts_json = os.path.join(
        config.project_dir, f"{series_name}", f"{series_name}_ts.json"
    )
    try:
        with open(ts_json, 'r') as tsj:
            config.terms_sheet = json.load(tsj, parse_float=decimal.Decimal)
    except FileNotFoundError:
        print(f"No Terms Sheet exists for Series {series_name} at {ts_json}")
        return ExitCode.EX_NOINPUT

    return config.terms_sheet


def load_assumed_collat() -> pd.DataFrame:
    """Load the Assumed Collateral replines from the Terms Sheet.

    This assumes that the Terms Sheet has already been loaded, using the
    ``load_deal()`` function.

    This will *only* load Assumed Collateral. PyMBS does not yet have the
    ability to handle Known Collateral or Securitzed Collateral. It can
    however handle multiple Assumed replines per group.
    """
    assumed_collat = core._load_assumed_collat()

    return assumed_collat


def load_model(model_json: TextIO) -> dict:
    """Load the structured cash flow model from the Terms Sheet, supplemented
    by the model file.

    The Terms sheet does NOT include the pay rule waterfall or specific
    benchmark interest rate information used by the model. These are provided
    in a separate JSON file.

    Within the project_dir, specified in the user's configuration file, it is
    expected that there will be one subdirectory for each REMIC Series name.
    Inside the Series directory, it is expected that the Model File named as
    provided as argument to this function will exist.

    Args:
        model_json: The name of the model file inside the Series directory.

    Returns:
        A dictionary object, representing the structured cash flow model.
        If the Model file is not located, returns Exit Code 66.

    Example:
        Given: ``model_json = 2618_model.json``

        Path: ``/project_dir/2618/2618_model.json``
    """
    if not config.terms_sheet:
        print(f"{exceptions.no_deal}")
        return ExitCode.EX_CONFIG
    else:
        ts = config.terms_sheet
    series_no = ts['meta']['series_no']
    model_file = os.path.join(
        config.project_dir, f"{series_no}", f"{series_no}_model.json"
    )
    try:
        with open(model_file, 'r') as tsj:
            mf = json.load(tsj, parse_float=decimal.Decimal)
    except FileNotFoundError:
        print(f"No model exists for Series {series_no} at {model_file}")
        return ExitCode.EX_NOINPUT
    groups = {}
    indices = {}
    for tranche in ts['tranches']['data']:
        tranche_group = tranche[0]
        if tranche_group not in groups:
            groups[tranche_group] = {}
    for group in groups:
        groups[group]['tranches'] = {}
        if group is not 0:
            groups[group]['collat_cf'] = {}
            groups[group]['waterfall'] = []
    first_payment_dates = ts['tranches']['first_payment_dates']
    for tranche in ts['tranches']['data']:
        tranche_group = tranche[0]
        tranche_id = tranche[1]
        groups[tranche_group]['tranches'][tranche_id] = Tranche(
            *tranche,
            dated_date=ts['meta']['issue_date'],
            next_payment_date=first_payment_dates[str(tranche_group)]
        )
    for group in groups:
        if group is not 0:
            groups[group]['waterfall'].extend(mf['waterfall'][str(group)])
    for index in mf['indices']:
        indices[index['name']] = IndexRate(**index)
    config.model['groups'] = groups
    config.model['indices'] = indices

    return config.model


def run_collat_cf(
    group_id: Optional[int] = -1,
    repline_num: Optional[int] = -1
) -> dict:
    """Run cash flows from Assumed Collateral replines specified in the
    Terms Sheet. A Pandas DataFrame is created for each cash flow prepayment
    scenario, as specified in the Prepayment Scenarios JSON file.

    The Terms sheet does NOT include the pay rule waterfall or specific
    benchmark interest rate information used by the model. These are provided
    in a separate JSON file.

    Within the project_dir, specified in the user's configuration file, it is
    expected that there will be one subdirectory for each REMIC Series name.
    Inside the Series directory, it is expected that the Model File named as
    provided as argument to this function will exist.

    Args:
        group_id: The number of the Group for which to run the cash flows.
                  By default, this value is set to -1, which will run
                  cash flows for ALL groups in the deal at the Prepayment
                  Scenarios specified for each group.

        repline_num: Optionally, the user may specify running the cash flows
                     for a specific repline in a specific group, if the group
                     has more than one assumed repline. As with ``group_id``,
                     the ``repline_num`` is set to -1 by default, which will
                     run cash flows for ALL replines associated with a group.
                     When run together, the cash flows for each repline will
                     be aggregated into one final cash flow which is the one
                     that is returned for the group.

    Returns:
        A dictionary object, containing a Pandas Dataframe for the cash flow
        calculated at each Prepayment Scenario.

    Example:
        Path: ``/project_dir/2618/2618_pps.json``
    """
    # TODO: Run collateral cash flows for known collateral and securitzed
    # collateral too!
    cash_flows = core._run_collat_cf(group_id, repline_num=0)

    return cash_flows


def show_wals(
    group_id: Optional[int] = -1,
    precision: Optional[int] = config.precision
) -> pd.DataFrame:
    """Calulate the Weigted Average Lives (WALs) of all Regular tranches in the
    group specified, including those for the collateral. Currently WALs are NOT
    calcuated for Notional or MACR classes. This functionality will be exposed
    in a future release.

    If no Group number is specified in the call to this function, WALs will
    be calculated for all groups in the deal. This is not necessarily
    desireable, as the Prepayment Scenarios will be different for different
    groups, so the table that is returned will contain a number of 'missing'
    values. An enhancement to handle this gracefully will be provided in a
    future release. In the meantime, it is recommended that the user specify
    a group number when calling this function.

    Args:
        group_id: The number of the Group for which to run the WALs.
                  By default, this value is set to -1, which will run
                  WALs for ALL groups in the deal at the Prepayment
                  Scenarios specified for each group.

        precision: The precision of the calculated WALs may be optionally
                   specified. If not specifed, it will use the
                   ``round_precision`` value specified in the configuration
                   object. This value is set to 10 decimal places by default,
                   which is almost always sufficient for tying-out the cash
                   flows with a counter-party. The precision of the WALs
                   disclosed in the Prospectus Supplement is 1 decimal.

    Returns:
        A Pandas Dataframe showing the WALs for each tranche, calulated based
        on the cash flows run at each Prepayment Sceanrio specified for the
        group.
    """
    if not config.model:
        print(f"{exceptions.no_deal}")
        return ExitCode.EX_CONFIG
    else:
        model = config.model
    wals = core._run_wals(group_id, precision, model)
    column_order = [wal for wal in wals['prepay_scenario'].unique()]

    def f(x):
        return (x.pivot('tranche_id', 'prepay_scenario', 'wal'))

    _wal_table = wals.groupby('group_id', group_keys=True).apply(f)
    wal_table = _wal_table.reindex(columns=column_order)

    return wal_table


def load_tranches(group_id: Optional[int] = -1) -> pd.DataFrame:
    """Load the tranches for the deal, or for the group specified, into a
    Pandas Dataframe for easy display in the Jupyter Notebook.

    Args:
        group_id: The number of the Group for which to load the tranches.
                  By default, this value is set to -1, which will load
                  the tranches for ALL of groups in the deal. As per
                  convention, the Residual tranches will appear in
                  Group 0.

    Returns:
        A Pandas Dataframe showing Tranche details for each tranche queried.
    """
    if not config.terms_sheet:
        print(f"{exceptions.no_deal}")
        return ExitCode.EX_CONFIG
    else:
        ts = config.terms_sheet

    tranches = pd.DataFrame.from_records(
        ts['tranches']['data'],
        columns=ts['tranches']['columns']
    )

    if group_id < 0:
        tranches = tranches
    else:
        tranches = tranches[tranches['group_id'] == group_id]

    tranches.index = range(1, len(tranches) + 1)

    return tranches


def disperse_cf(model: dict, group_id: Optional[int] = -1) -> int:
    """Make all payments of Interest and Principal, based on the rules
    enumerated in the model.

    Args:
        group_id: The number of the Group for which to load the tranches.
                  By default, this value is set to -1, which will load
                  the tranches for ALL of groups in the deal. As per
                  convention, the Residual tranches will appear in
                  Group 0.

    Returns:
        ExitCode 0 on success. After this function has returned, the cash flows
        for each tranche, including the collateral, will be accessible from
        within the model.
    """
    if group_id < 0:
        for group in model['groups']:
            cash_flows = run_collat_cf(group)
            model['groups'][group]['collat_cf'].update(cash_flows)
            tranches = core._get_regular_tranches(model, group)
            core._make_payments(model, group, cash_flows, tranches)
    else:
        cash_flows = run_collat_cf(group_id)
        model['groups'][group_id]['collat_cf'].update(cash_flows)
        tranches = core._get_regular_tranches(model, group_id)
        core._make_payments(model, group_id, cash_flows, tranches)

    return ExitCode.EX_SUCCESS



"""
PyMBS is a Python library for use in modeling Mortgage-Backed Securities.

Copyright (C) 2019  Brian Farrell

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact: brian.farrell@me.com
"""

from enum import Enum, IntEnum


class ExitCode(IntEnum):
    """Enumeration of exit codes.

    This class implements POSIX error codes, as found in
    /usr/include/sysexits.h on the Linux platform and
    described in the manual page for sysexit on BSD systems,
    including Darwin (macOS).

    Additional information is availble from The Linux Documnetation Project
    at http://tldp.org/LDP/abs/html/exitcodes.html
    """
    EX_SUCCESS = 0  # command exits successfully
    EX_GENERAL = 1  # catchall for general errors
    EX_USAGE = 64  # command line usage error
    EX_DATAERR = 65  # data format error
    EX_NOINPUT = 66  # cannot open input
    EX_NOUSER = 67  # addressee unknown
    EX_NOHOST = 68  # host name unknown
    EX_UNAVAILABLE = 69  # service unavailable
    EX_SOFTWARE = 70  # internal software error
    EX_OSERR = 71  # system error (e.g., can't fork)
    EX_OSFILE = 72  # critical OS file missing
    EX_CANTCREAT = 73  # can't create (user) output file
    EX_IOERR = 74  # input/output error
    EX_TEMPFAIL = 75  # temp failure; user is invited to retry
    EX_PROTOCOL = 76  # remote error in protocol
    EX_NOPERM = 77  # permission denied
    EX_CONFIG = 78  # configuration error
    EX_SIGINT = 130  # keyboard interrupt


class PrepayBenchmark(Enum):
    """docstring for PrepayBenchmark"""
    CPR = 'CPR'
    PSA = 'PSA'



"""
PyMBS is a Python library for use in modeling Mortgage-Backed Securities.

Copyright (C) 2019  Brian Farrell

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact: brian.farrell@me.com
"""

from setuptools import setup, find_packages

from pymbs.__version__ import __version__, _version_min_python_

with open("README.rst", "r") as fh:
    long_description = fh.read()

required = [
    "ipython>=7.7.0",
    "jinja2>=2.10.1",
    "numpy>=1.16.4",
    "pandas>=0.25.0",
    "pyyaml>=5.1.2",
]

setup(
    name='pymbs',
    version=__version__,
    python_requires=f">={_version_min_python_}",
    install_requires=required,

    packages=find_packages(),
    data_files=[
        ('pymbs/config', ['pymbs/config/pymbs_logging.json']),
        ('pymbs/templates', ['pymbs/templates/new_config.html'])
    ],

    license='AGPLv3',
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Financial and Insurance Industry",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "License :: OSI Approved :: "
        "GNU Affero General Public License v3 or later (AGPLv3+)",
        "Topic :: Office/Business :: Financial",
        "Topic :: Office/Business :: Financial :: Investment",
        "Topic :: Software Development :: Libraries",
        "Operating System :: POSIX",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: CPython",
    ],

    url='https://pypi.org/project/pymbs/',
    author="Brian Farrell",
    author_email="brian.farrell@me.com",
    description="A library for use in modeling Mortgage-Backed Securities.",
    long_description=long_description,
    long_description_content_type='text/x-rst',
    keywords="financial modeling analysis mortgage securities",
    zip_safe=False,
)



"""
PyMBS is a Python library for use in modeling Mortgage-Backed Securities.

Copyright (C) 2019  Brian Farrell

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact: brian.farrell@me.com

The log module reads-in the pymbs_logging.json file and sets up logging for
PyMBS.

The PyMBS log file is located at ~/.local/log/pymbs.log

"""

import json
import logging
import logging.config
from pathlib import Path
import os
import sys

from pymbs.enums import ExitCode

platform = sys.platform

PYMBS_BASE_PATH = Path(__file__).parent

log_dir = 'pymbs'
log_file = 'pymbs.log'
rel_log_path = Path(log_dir, log_file)
user_home_dir = Path.home()
log_dir_mode = 0o774

if platform == 'darwin':
    pymbs_log_dir = Path(user_home_dir, '.local', 'log', log_dir)
elif platform == 'linux':
    pymbs_log_dir = Path('var', 'log', log_dir)
elif platform == 'win32':
    pymbs_log_dir = Path(user_home_dir, 'AppData', 'Local', log_dir)

if not pymbs_log_dir.exists():
    pymbs_log_dir.mkdir(mode=log_dir_mode, parents=True)

PYMBS_LOG_FILE = Path(pymbs_log_dir, log_file)

if not PYMBS_LOG_FILE.exists():
    PYMBS_LOG_FILE.touch()

PYMBS_LOG_CONFIG = Path(PYMBS_BASE_PATH, 'config', 'pymbs_logging.json')

PYMBS_LOG_LEVEL = 'INFO'


def _get_logging_config(env_log_level='PYMBS_LOG_LEVEL',
                        env_log_config='PYMBS_LOG_CONFIG',
                        env_log_file='PYMBS_LOG_FILE'):
    """Setup logging configuration."""
    value = os.getenv(env_log_config)
    if value:
        log_config_path = value
    else:
        log_config_path = PYMBS_LOG_CONFIG

    try:
        with open(log_config_path, 'rt') as f:
            logging_config = json.load(f)
    except OSError:
        print(f"Failed to open PyMBS logging configuration file at "
              f"{log_config_path}\nExiting...\n\n")
        sys.exit(ExitCode.EX_NOINPUT)

    value = os.getenv(env_log_file)
    if value:
        pymbs_log_file = value
    else:
        pymbs_log_file = PYMBS_LOG_FILE
    logging_config['handlers']['default_handler']['filename'] = pymbs_log_file

    value = os.getenv(env_log_level)
    if value:
        pymbs_log_level = value
    else:
        pymbs_log_level = PYMBS_LOG_LEVEL
    logging_config['loggers']['pymbs']['level'] = pymbs_log_level

    return logging_config


logging.config.dictConfig(_get_logging_config())


def get_logger(logger_name):
    logger = logging.getLogger(logger_name)

    return logger



"""
PyMBS is a Python library for use in modeling Mortgage-Backed Securities.

Copyright (C) 2019  Brian Farrell

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact: brian.farrell@me.com
"""

import os
from pathlib import Path
import sys

from IPython.display import display, HTML
from jinja2 import Environment, PackageLoader, select_autoescape
import yaml

from pymbs.enums import ExitCode
from pymbs.exceptions import _no_config_msg
from pymbs.log import get_logger

logger = get_logger(__name__)

# Python Decimal Module
DEFAULT_PRECISION = 18
DEFAULT_EMAX = 12
DEFAULT_EMIN = -10
DEFAULT_ROUND_PRECISION = 10

# Pandas
DEFAULT_MAX_ROWS = 400


class Config(object):
    """The Config object holds all configuration values for PyMBS, except for
    those related to Logging, which are stored in the ``pymbs.log`` module.
    See the documentation for that module in regards to the Logging settings.

    The user is able to customize settings via the ``config.yaml`` file, which
    is located in a subdirectory of the user's HOME directory, and is
    described further in the :ref:`configuration` section of this
    documentation.

    Many of the settings in the config object *may* be specifed in
    environment variables. If a value for a setting is specified simultaneously
    in the ``config.yaml`` file *and* as an environment variable, the
    value from the environment variable will be used.

    As noted below, some of the properties set in the ``config`` object are
    actually attributes of the ``Context`` object in Python's ``Decimal``
    module.

    For more information on the Decimal Context object, see:

    https://docs.python.org/3/library/decimal.html#decimal.Context

    Some of the properties set in the ``config`` object are actually options
    in the ``Pandas`` library.

    For more information, see:

    https://pandas.pydata.org/pandas-docs/stable/user_guide/options.html#available-options
    """
    _platform = sys.platform
    _config_dir = 'pymbs'
    _config_file = 'config.yaml'
    _rel_config_path = Path(_config_dir, _config_file)
    _user_home_dir = Path.home()
    _config_file_mode = 0o774
    _help_url = 'https://brianfarrell.gitlab.io/pymbs/configuration.html'

    if _platform != 'win32':
        _user_app_config_dir = Path(_user_home_dir, '.config')
    else:
        _user_app_config_dir = Path(_user_home_dir, 'AppData', 'Local')

    _pymbs_config_path = Path(
        _user_app_config_dir, _rel_config_path
    )

    _jinja_env = Environment(
        loader=PackageLoader('pymbs', 'templates'),
        autoescape=select_autoescape(['html', 'xml'])
    )

    _ipython_active = False

    def __init__(self):
        Config._detect_ipython()
        _config = Config._load_config()

        # PyMBS Config
        self._project_dir = None
        self._terms_sheet = None
        self._model = {}
        self._cache = {}

        # Pandas Config
        self._precision = None
        self._emax = None
        self._emin = None
        self._round_precision = None
        self._max_rows = None

        self._configure(_config)

    ########################
    ##### PyMBS Config #####
    ########################

    @property
    def project_dir(self):
        """The project_dir *must* be set, as it can not be known or
        discovered in advance.

        The project directory is a dicrecotry in the file system that contains
        one subdirectory for each deal modeled with PyMBS. By convention,
        each subdirectory is named using the name of the deal being modeled.
        For more information about the directory structure, see the
        :ref:`setup` section of this documentation.

        This may set in the ``config.yaml`` file, or via the
        ``PYMBS_PROJECT_DIR`` environment variable.
        """
        return self._project_dir

    @project_dir.setter
    def project_dir(self, value):
        logger.info(f"Setting project_dir in config: {value}")
        self._project_dir = value

    @property
    def terms_sheet(self):
        """The terms_sheet value should *never* be set directly. This value
        is set by calling the ``api.load_deal(series_name)`` function,
        passing the name of the directory that holds the files for the deal.

        The ``load_deal`` function deserializes the ``series_ts.json`` file
        and converts all ``Float`` values into ``Decimal`` values, using
        Python's Decimal module.

        The Terms Sheet structure may be viewed as a Python dictionary by
        referencing the ``config.terms_sheet`` property.
        """
        return self._terms_sheet

    @terms_sheet.setter
    def terms_sheet(self, value):
        logger.info(
            f"Setting terms_sheet in config: "
            f"Series: {value['meta']['series_no']} | "
            f"TS Version: {value['meta']['ts_version']} | "
            f"Underwriter: {value['meta']['lead_underwriter']}"
        )
        self._terms_sheet = value

    @property
    def model(self):
        """The model value should *never* be set directly. This value
        is set by calling the ``api.load_model(model_json)`` function,
        passing the name of the JSON file that holds the model for the deal.

        The ``load_model`` function deserializes the JSON data into Python
        data types, for use in all subsequent function calls to compute the
        outputs of the model.

        The model structure may be viewed as a Python dictionary by
        referencing the ``config.model`` property.
        """
        return self._model

    @property
    def cache(self):
        """The cache value should *never* be set directly. This Python
        dictionary is referenced and modified by private functions throughout
        the PyMBS code.

        The values in this dictionary change quicky when running the model, so
        viewing it by referencing the ``config.cache`` property may not make
        sense, except in a debugging context.
        """
        return self._cache

    @property
    def round_precision(self):
        """round_precision is the number of decimals to round to, when
        rounding a decimal value. It is *not* part of the specification
        described in the Python Decimal Module, but is a custom value
        created and used by PyMBS.

        This value is used in the custom ``_round_dec()`` function, located
        in the ``pymbs.utils`` module.

        This may set in the ``config.yaml`` file, or via the
        ``PYMBS_ROUND_PRECISION`` environment variable.
        """
        return self._round_precision

    @round_precision.setter
    def round_precision(self, value):
        logger.info(f"Setting round_precision in config: {value}")
        self._round_precision = value

    #################################
    ##### Python Decimal Config #####
    #################################

    @property
    def precision(self):
        """prec[ision] is an attribute of the ``Context`` object in Python's
        Decimal module.

        The precision is an integer in the range [1, MAX_PREC]
        that sets the precision for arithmetic operations in the context.

        This may set in the ``config.yaml`` file, or via the
        ``PYMBS_PRECISION`` environment variable.
        """
        return self._precision

    @precision.setter
    def precision(self, value):
        logger.info(f"Setting precision in config: {value}")
        self._precision = value

    @property
    def emax(self):
        """Emax is an attribute of the ``Context`` object in Python's
        Decimal module. The Emax field is an integer specifying the outer
        limit allowable for the max exponent.

        Emax must be in the range [0, MAX_EMAX].

        This may set in the ``config.yaml`` file, or via the ``PYMBS_EMAX``
        environment variable.
        """
        return self._emax

    @emax.setter
    def emax(self, value):
        logger.info(f"Setting emax in config: {value}")
        self._emax = value

    @property
    def emin(self):
        """Emin is an attribute of the ``Context`` object in Python's
        Decimal module. The Emin field is an integer specifying the outer
        limit allowable for the min exponent.

        Emin must be in the range [MIN_EMIN, 0].

        This may set in the ``config.yaml`` file, or via the ``PYMBS_EMIN``
        environment variable.
        """
        return self._emin

    @emin.setter
    def emin(self, value):
        logger.info(f"Setting emin in config: {value}")
        self._emin = value

    #########################
    ##### Pandas Config #####
    #########################

    @property
    def max_rows(self):
        """This sets the maximum number of rows Pandas should output when
        printing out various output. For example, this value determines
        whether the repr() for a dataframe prints out fully or just a
        truncated or summary repr. ‘None’ value means unlimited.

        The Pandas default is 60.

        The PyMBS default for this value is 400, which was determined by
        rounding up from 360, which is the number of monthly periods in a
        30-year cash flow.

        This may set in the ``config.yaml`` file, or via the ``PYMBS_MAX_ROWS``
        environment variable.
        """
        return self._max_rows

    @max_rows.setter
    def max_rows(self, value):
        logger.info(f"Setting max_rows in config: {value}")
        self._max_rows = value

    def _configure(self, _config):
        _env_project_dir = os.getenv('PYMBS_PROJECT_DIR')

        _env_precision = os.getenv('PYMBS_PRECISION')
        _env_emax = os.getenv('PYMBS_EMAX')
        _env_emin = os.getenv('PYMBS_EMIN')
        _env_round_precision = os.getenv('PYMBS_ROUND_PRECISION')
        _env_max_rows = os.getenv('PYMBS_MAX_ROWS')

        _file_project_dir = None

        _file_precision = None
        _file_emax = None
        _file_emin = None
        _file_round_precision = None
        _file_max_rows = None

        pymbs_config = _config.get('pymbs')
        if pymbs_config:
            _file_project_dir = pymbs_config.get('project directory')

        pandas_config = _config.get('pandas')
        if pandas_config:
            _file_precision = pandas_config.get('precision')
            _file_emax = pandas_config.get('emax')
            _file_emin = pandas_config.get('emin')
            _file_round_precision = pandas_config.get('round precision')
            _file_max_rows = pandas_config.get('max rows')

        if _env_project_dir and _file_project_dir:
            self.project_dir = _env_project_dir
        elif _env_project_dir:
            self.project_dir = _env_project_dir
        elif _file_project_dir:
            self.project_dir = _file_project_dir

        if _env_precision and _file_precision:
            self.precision = _env_precision
        elif _env_precision:
            self.precision = _env_precision
        elif _file_precision:
            self.precision = _file_precision
        else:
            self.precision = DEFAULT_PRECISION

        if _env_emax and _file_emax:
            self.emax = _env_emax
        elif _env_emax:
            self.emax = _env_emax
        elif _file_emax:
            self.emax = _file_emax
        else:
            self.emax = DEFAULT_EMAX

        if _env_emin and _file_emin:
            self.emin = _env_emin
        elif _env_emin:
            self.emin = _env_emin
        elif _file_emin:
            self.emin = _file_emin
        else:
            self.emin = DEFAULT_EMIN

        if _env_round_precision and _file_round_precision:
            self.round_precision = _env_round_precision
        elif _env_round_precision:
            self.round_precision = _env_round_precision
        elif _file_round_precision:
            self.round_precision = _file_round_precision
        else:
            self.round_precision = DEFAULT_ROUND_PRECISION

        if _env_max_rows and _file_max_rows:
            self.max_rows = _env_max_rows
        elif _env_max_rows:
            self.max_rows = _env_max_rows
        elif _file_max_rows:
            self.max_rows = _file_max_rows
        else:
            self.max_rows = DEFAULT_MAX_ROWS

    @classmethod
    def _detect_ipython(cls):
        try:
            cfg = get_ipython().config  # noqa
            cls._ipython_active = True
        except NameError:
            cls._ipython_active = False

    @staticmethod
    def _load_config():
        alt_config_path = os.getenv('PYMBS_CONFIG_PATH')
        if alt_config_path:
            _pymbs_config_path = alt_config_path
        else:
            _pymbs_config_path = Config._pymbs_config_path
        try:
            with open(_pymbs_config_path, 'r') as stream:
                data = yaml.full_load(stream)
        except FileNotFoundError:
            pymbs_config_dir = Path(
                Config._user_app_config_dir, Config._config_dir
            )
            if not pymbs_config_dir.exists():
                pymbs_config_dir.mkdir(
                    mode=Config._config_file_mode, parents=True
                )
            _pymbs_config_path.touch()

            if not Config._ipython_active:
                print(_no_config_msg(_pymbs_config_path, Config._help_url))
                sys.exit(ExitCode.EX_CONFIG)
            else:
                no_config_html = Config._jinja_env.get_template(
                    'new_config.html'
                )
                msg = no_config_html.render(
                    _pymbs_config_path=_pymbs_config_path,
                    _help_url=Config._help_url
                )
                display(HTML(msg))
        else:
            return data


config = Config()



"""
PyMBS is a Python library for use in modeling Mortgage-Backed Securities.
Copyright (C) 2019  Brian Farrell

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact: brian.farrell@me.com
"""

from invoke import task
import pytest


@task
def full(ctx):
    """
    Run The Full Test Suite
    """
    pytest.main(['tests'])


@task
def unit(ctx):
    """
    Run Unit Tests
    """
    pytest.main(['tests'])


@task
def func(ctx):
    """
    Run Functional Tests
    """
    pytest.main(['tests'])


@task
def sub(ctx):
    """
    Run Subcutaneous Tests
    """
    pytest.main(['tests'])
